/*****************************************************************************/
/* MFCMIDI.CPP : class implementation for MIDI base classes                  */
/*****************************************************************************/
/*                                                                           */
/* $Archive::                                                              $ */
/* $Revision::                                                             $ */
/* $Date::                                                                 $ */
/* $Author::                                                               $ */
/* $Modtime::                                                              $ */
/* $Workfile::                                                             $ */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Includes                                                                  */
/*****************************************************************************/

#include "stdafx.h"                     /* MFC include                       */
#include "resource.h"                   /* Project resources                 */
#include "mfcmidi.h"                    /* private definitions               */

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*****************************************************************************/
/* Definitions                                                               */
/*****************************************************************************/

#define MIDIINHDRS      7               /* # MIDI SysEx Input headers        */
#define MIDIINBUFSIZE   8192L           /* length of MIDI input buffer       */

/*****************************************************************************/
/* Global Data                                                               */
/*****************************************************************************/

static char szMMapName[]="MIDI Mapper"; /* MIDI Mapper device name           */

/*===========================================================================*/
/* Class CMidiMsg                                                            */
/*===========================================================================*/

/*****************************************************************************/
/* CMidiMsg : constructor                                                    */
/*****************************************************************************/

CMidiMsg::CMidiMsg
    (
    DWORD smsg
    )
{
Init();
Set(smsg);
}

CMidiMsg::CMidiMsg
    (
    int b1,
    int b2,
    int b3
    )
{
DWORD smsg = 0;

Init();
Set(b1, b2, b3);
}

CMidiMsg::CMidiMsg
    (
    LPMIDIHDR lphdr
    )
{
Init();
Set(lphdr);
}

CMidiMsg::CMidiMsg
    (
    LPSTR data,
    int len
    )
{
Init();
Set(data,len);
}

/*****************************************************************************/
/* Init : initializes CMidiMsg storage                                       */
/*****************************************************************************/

void CMidiMsg::Init ( )
{
bAlloc = FALSE;
sMsg = 0;
lpHdr = NULL;
hGlPtr = NULL;
dwStamp = 0;
}

/*****************************************************************************/
/* Empty : empties preinitialized CMidiMsg storage                           */
/*****************************************************************************/

void CMidiMsg::Empty ( )
{
if (bAlloc)
  {
  GlobalUnlock(hGlPtr);
  GlobalFree(hGlPtr);
  }
Init();
}

/*****************************************************************************/
/* Set : sets to MIDI message of various types                               */
/*****************************************************************************/

void CMidiMsg::Set
    (
    int b1,
    int b2,
    int b3
    )
{
DWORD smsg =
    (DWORD)(b1 & 0xffL) +
    (DWORD)((b2 & 0xffL) << 8L) +
    (DWORD)((b3 & 0xffL) << 16L);
Set(smsg);
}

void CMidiMsg::Set
    (
    DWORD smsg
    )
{
Empty();
sMsg = smsg;
if (((sMsg & 0xf0) == 0x90) &&          /* if NOTE ON that's really NOTE OFF */
    (!(sMsg & 0xff0000L)))
                                        /* make it a NOTE OFF with vel.64    */
  sMsg = (sMsg & (0x00ff0fL)) | 0x400080L;     
}

void CMidiMsg::Set
    (
    LPMIDIHDR lphdr
    )
{
Empty();
lpHdr = lphdr;
}

void CMidiMsg::Set
    (
    LPSTR data,
    int len
    )
{
if (len <= 3)
  {
  Set(*((DWORD *)data));
  return;
  }

Empty();

hGlPtr = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,
    sizeof(MIDIHDR) + len);
if (!hGlPtr)
  return;
lpHdr = (LPMIDIHDR)GlobalLock(hGlPtr);
if (lpHdr)
  {
  bAlloc = TRUE;
  memset((LPVOID)lpHdr, 0, sizeof(MIDIHDR));
  lpHdr->lpData = (LPSTR)(lpHdr+1);
  lpHdr->dwBufferLength = (DWORD)len;
  lpHdr->dwBytesRecorded = (DWORD)len;
  memcpy(lpHdr->lpData, data, len);
  }
}

/*****************************************************************************/
/* operator char * : returns the message as a character pointer              */
/*****************************************************************************/

CMidiMsg::operator char * ( )
{
if (!lpHdr)                             /* if no header                      */
  return (char *)&sMsg;                 /* pass back short MIDI msg address  */
return (char *)lpHdr->lpData;           /* otherwise pass back data pointer  */
}

/*****************************************************************************/
/* operator BYTE * : returns the message as a character pointer              */
/*****************************************************************************/

CMidiMsg::operator BYTE * ( )
{
if (!lpHdr)                             /* if no header                      */
  return (BYTE *)&sMsg;                 /* pass back short MIDI msg address  */
return (BYTE *)lpHdr->lpData;           /* otherwise pass back data pointer  */
}

/*****************************************************************************/
/* operator DWORD : returns short message as DWORD                           */
/*****************************************************************************/

CMidiMsg::operator DWORD ( )
{
if (!lpHdr)                             /* if no header                      */
  return sMsg;                          /* pass back short MIDI msg          */
return *((DWORD *)lpHdr->lpData);       /* otherwise pass back data pointer  */
}

/*****************************************************************************/
/* operator LPMIDIHDR : returns the message as a MIDIHDR pointer             */
/*****************************************************************************/

CMidiMsg::operator LPMIDIHDR ( )
{
return lpHdr;                           /* pass back MIDI header             */
}

/*****************************************************************************/
/* Length : returns the length of the passed message                         */
/*****************************************************************************/

WORD CMidiMsg::Length ( )
{
if (!lpHdr)                             /* if a short message                */
  {
  switch (sMsg & 0xf0)                  /* filter all known lengths          */
    {
    case STATUS_PROGRAMCHANGE :         /* 2-byte messages                   */
    case STATUS_CHANNELPRESSURE  :
      return 2;
    case STATUS_NOTEOFF :               /* 3-byte messages                   */
    case STATUS_NOTEON :
    case STATUS_POLYPHONICKEY :
    case STATUS_CONTROLCHANGE :
    case STATUS_PITCHBEND :
      return 3;
    }

  switch (sMsg & 0xff)                  /* now process the others            */
    {
    case STATUS_TUNEREQUEST :           /* 1-byte codes                      */
    case STATUS_TIMINGCLOCK :
    case STATUS_START :
    case STATUS_CONTINUE :
    case STATUS_STOP :
    case STATUS_ACTIVESENSING :
    case STATUS_SYSTEMRESET :
      return 1;
    case STATUS_SONGSELECT :            /* 2-byte codes                      */
      return 2;
    case STATUS_QFRAME :                /* 3-byte codes                      */
    case STATUS_SONGPOINTER :
      return 3;
    }

  return 0;                             /* return invalid                    */
  }

return (WORD)                           /* otherwise return length of buffer */
    (lpHdr->dwBytesRecorded ?
        lpHdr->dwBytesRecorded :
        lpHdr->dwBufferLength);
}

/*===========================================================================*/
/* Class CMidiDevice                                                         */
/*===========================================================================*/

/*****************************************************************************/
/* CMidiDevice : constructor                                                 */
/*****************************************************************************/

CMidiDevice::CMidiDevice ( )
{
szDevName[0] = 0;                       /* reset device name                 */
bIsOpen = FALSE;                        /* reset opened flag                 */
}

/*****************************************************************************/
/* ~CMidiDevice : destructor                                                 */
/*****************************************************************************/

CMidiDevice::~CMidiDevice ( )
{
}

/*****************************************************************************/
/* CMidiDevice message map                                                   */
/*****************************************************************************/

BEGIN_MESSAGE_MAP(CMidiDevice, CWnd)
	//{{AFX_MSG_MAP(CMidiDevice)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*****************************************************************************/
/* Create : creates necessary notification window                            */
/*****************************************************************************/

BOOL CMidiDevice::Create ( )
{
if (!CreateEx(0,
    AfxRegisterWndClass(0),
    "Midi Wnd",
    WS_POPUP,
    0,
    0,
    0,
    0,
    NULL,
    NULL))
  {
  TRACE("Failed to create midi notification window");
  return FALSE;
  }
return TRUE;
}

/*****************************************************************************/
/* Open : opens the device                                                   */
/*****************************************************************************/

BOOL CMidiDevice::Open
    (
    int iID,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
AssureWnd();                            /* make sure window is there         */
bIsOpen = TRUE;
OnOpen();
return TRUE;
}

BOOL CMidiDevice::Open
    (
    LPCSTR szName,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
strcpy(szDevName, szName);
return TRUE;
}

/*****************************************************************************/
/* Close : closes the device                                                 */
/*****************************************************************************/

BOOL CMidiDevice::Close ( )
{
*szDevName = 0;
bIsOpen = FALSE;
OnClose();
return TRUE;
}

/*****************************************************************************/
/* Reset : resets the attached MIDI device                                   */
/*****************************************************************************/

BOOL CMidiDevice::Reset ( )
{
return TRUE;                            /* pass back OK                      */
}

/*===========================================================================*/
/* Class CMidiInDeviceList                                                   */
/*===========================================================================*/

/*****************************************************************************/
/* CMidiInDeviceList : constructor                                           */
/*****************************************************************************/

CMidiInDeviceList::CMidiInDeviceList ( )
{
WORD wNumDevices;
int i;
MIDIINCAPS mcap;

wNumDevices = midiInGetNumDevs();       /* get # MIDI input devices          */
if (wNumDevices)
  {
  for (i=0; i < (int) wNumDevices; i++)
    {
    midiInGetDevCaps(i, &mcap, sizeof(mcap));
    Add(mcap.szPname);
    }
  }
}

/*===========================================================================*/
/* Class CMidiInDevice                                                       */
/*===========================================================================*/

/*****************************************************************************/
/* CMidiInDevice message map                                                 */
/*****************************************************************************/

BEGIN_MESSAGE_MAP(CMidiInDevice, CMidiDevice)
    //{{AFX_MSG_MAP(CMidiInDevice)
    ON_MESSAGE(MM_MIM_OPEN, OnMimOpen)
    ON_MESSAGE(MM_MIM_CLOSE, OnMimClose)
    ON_MESSAGE(MM_MIM_ERROR, OnMimError)
    ON_MESSAGE(MM_MIM_LONGERROR, OnMimLongError)
    ON_MESSAGE(MM_MIM_DATA, OnMimData)
    ON_MESSAGE(MM_MIM_LONGDATA, OnMimLongData)
    ON_MESSAGE(MM_MIM_MOREDATA, OnMimMoreData)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*****************************************************************************/
/* CMidiInDevice : constructor                                               */
/*****************************************************************************/

CMidiInDevice::CMidiInDevice ( )
{
hDev = NULL;                            /* reset device handle               */

hGlobal = NULL;                         /* MIDI In buffers                   */
pGlobal = NULL;
lpGlHdr = new LPMIDIHDR[MIDIINHDRS];
if (lpGlHdr)
  for (int i = 0; i < MIDIINHDRS; i++)
    lpGlHdr[i] = NULL;
bRecording = FALSE;                     /* reset recording flag              */
dwStamp = 0;                            /* set internal time stamp           */
bClosing = FALSE;
}

/*****************************************************************************/
/* ~CMidiInDevice : destructor                                               */
/*****************************************************************************/

CMidiInDevice::~CMidiInDevice ( )
{
if (IsOpen())
  Close();

FreeBuf();
delete[] lpGlHdr;
}

/*****************************************************************************/
/* MidiInProc : callback procedure                                           */
/*****************************************************************************/

void CALLBACK CMidiInDevice::MidiInProc
    (
    HMIDIIN hMidiIn,
    UINT wMsg,
    DWORD dwInstance,
    DWORD dwParam1,
    DWORD dwParam2
    )
{
CMidiInDevice * pInDev = (CMidiInDevice *)dwInstance;
switch (wMsg)
  {
  case MIM_CLOSE : 
    pInDev->OnMimClose();
    break;
  case MIM_OPEN : 
    pInDev->OnMimOpen();
    break;
  case MIM_ERROR :
    {
    CMidiMsg iMsg(dwParam1);            /* generate MIDI message             */
    iMsg.SetStamp(dwParam2);
    pInDev->OnMimError(iMsg);
    }
    break;
  case MIM_LONGERROR : 
    pInDev->OnMimError((LPMIDIHDR)dwParam1);
    break;
  case MIM_MOREDATA :
    {
    CMidiMsg iMsg(dwParam1);            /* generate MIDI message             */
    iMsg.SetStamp(dwParam2);
    pInDev->OnMimMoreData(iMsg);
    }
    break;
  case MIM_DATA : 
    {
    CMidiMsg iMsg(dwParam1);            /* generate MIDI message             */
    iMsg.SetStamp(dwParam2);
    pInDev->Data(iMsg);                 /* process the message               */
    }
    break;
  case MIM_LONGDATA : 
                                        /* process the message               */
    pInDev->OnMimLongData((LPMIDIHDR)dwParam1, dwParam2);
    break;
  }
}

/*****************************************************************************/
/* Open : opens a MIDI In Device with its ID                                 */
/*****************************************************************************/

BOOL CMidiInDevice::Open
    (
    int iID,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
if (IsOpen())                           /* if already open                   */
  Close();                              /* close the opened device           */

AssureWnd();                            /* make sure window is there         */

if ((dwFlags & CALLBACK_TYPEMASK) ==    /* if function callback              */
    CALLBACK_FUNCTION)
  {
  if (lp1 == OPENDEFAULT)               /* if default value                  */
    lp1 = (DWORD)MidiInProc;            /* set LP1 to procedure              */
  if (lp2 == OPENDEFAULT)               /* if default value                  */
    lp2 = (DWORD)this;                  /* set LP2 to object                 */
  }
                                        /* if thread callback                */
else if ((dwFlags & CALLBACK_TYPEMASK) ==    
    CALLBACK_THREAD)
  {
  if (lp1 == OPENDEFAULT)               /* if default value                  */
    lp1 = (DWORD)GetCurrentThreadId();  /* set LP1 to thread ID              */
  if (lp2 == OPENDEFAULT)               /* if default value                  */
    lp2 = (UINT)AfxGetInstanceHandle(); /* set LP2 to instance               */
  }
else
  {
  if (lp1 == OPENDEFAULT)               /* if default value                  */
    lp1 = (UINT)m_hWnd;                 /* set LP1 to window                 */
  if (lp2 == OPENDEFAULT)               /* if default value                  */
    lp2 = (UINT)AfxGetInstanceHandle(); /* set LP2 to instance               */
  }

WORD wRtn = 1;
try
  {
  wRtn = midiInOpen(&hDev,              /* open the device                   */
      iID,                              /* with the passed ID                */
      lp1, lp2, dwFlags);
  }
catch(...)
  {
  }
if (wRtn)                               /* if open error                     */
  {
  hDev = NULL;                          /* make sure that handle is reset    */
  return FALSE;
  }

if (!AllocBuf())                        /* allocate MIDI Input buffers       */
  {
  Close();                              /* upon error close device           */
  return FALSE;                         /* and return NOT GOOD               */
  }

                                        /* call base class                   */
CMidiDevice::Open(iID, lp1, lp2, dwFlags);

midiInGetDevCaps(iID,                   /* get device capabilities           */
    &mCap, sizeof(mCap));

#if 0
return Start();                         /* start recording                   */
#else
return TRUE;
#endif                                  /* return OK                         */
}

/*****************************************************************************/
/* Open : opens a device with its name                                       */
/*****************************************************************************/

BOOL CMidiInDevice::Open
    (
    LPCSTR szName,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
WORD wNumDevices;
int i;
WORD wRtn;
MIDIINCAPS mcap;

wNumDevices = midiInGetNumDevs();       /* get # MIDI input devices          */
if (wNumDevices)
  {
  for (i=-1; i < (int) wNumDevices; i++)
    {
    if (i == -1)
      lstrcpy(mcap.szPname, szMMapName);
    else
      wRtn = midiInGetDevCaps(i, &mcap, sizeof(mcap));
    if (!lstrcmp(szName, mcap.szPname))
      {
      if (!Open(i, lp1, lp2, dwFlags))
        return(FALSE);                  /* error - return NOT OK             */

      return CMidiDevice::Open(szName,  /* opened - return OK                */
          lp1, lp2, dwFlags);
      }
    }
  }
return(FALSE);                          /* return NOT OK                     */
}

/*****************************************************************************/
/* Start : start recording from device                                       */
/*****************************************************************************/

BOOL CMidiInDevice::Start ( )
{
if (!IsOpen() || IsRecording())         /* if no device open or recording    */
  return FALSE;                         /* return error                      */

if ((!PrepareBuf()) ||                  /* start recording from device       */
    (midiInStart(hDev)))
  return FALSE;                         /* upon error return error           */
dwStamp = timeGetTime();                /* set internal time stamp           */

bRecording = TRUE;                      /* set recording flag                */
return TRUE;
}

/*****************************************************************************/
/* Stop : stop recording from device                                         */
/*****************************************************************************/

BOOL CMidiInDevice::Stop ( )
{
if (!IsOpen() || !IsRecording())        /* if no device open or recording    */
  return FALSE;                         /* return error                      */

bRecording = FALSE;                     /* reset recording flag              */
Sleep(10);                              /* allow others to find that out     */
midiInStop(hDev);                       /* stop recording                    */
dwStamp = 0;                            /* set internal time stamp           */
return TRUE;
}

/*****************************************************************************/
/* PrepareHeader : prepares a header for use in mmsystem                     */
/*****************************************************************************/

BOOL CMidiInDevice::PrepareHeader(LPMIDIHDR lpHdr)
{
if (!hDev)
  return FALSE;
return !midiInPrepareHeader(hDev, lpHdr, sizeof(MIDIHDR));
}

/*****************************************************************************/
/* UnprepareHeader : unprepares a used header                                */
/*****************************************************************************/

BOOL CMidiInDevice::UnprepareHeader(LPMIDIHDR lpHdr)
{
if (!hDev)
  return FALSE;
return !midiInUnprepareHeader(hDev, lpHdr, sizeof(MIDIHDR));
}

/*****************************************************************************/
/* Reset : resets the attached MIDI In device                                */
/*****************************************************************************/

BOOL CMidiInDevice::Reset ( )
{
if (!IsOpen())                          /* if no device open                 */
  return FALSE;                         /* return error                      */
Stop();                                 /* make sure recording's stopped     */
midiInReset(hDev);                      /* reset the attached device         */
while (PreparedBuffers())               /* wait until all buffers unprepared */
  {
  TRACE1("CMidiInDevice::Reset(): waiting for %d buffers to be unprepared\n",
         PreparedBuffers());
  Sleep(10);
  }
return TRUE;                            /* and pass back OK                  */
}

/*****************************************************************************/
/* Close : closes the opened MIDI In Device                                  */
/*****************************************************************************/

BOOL CMidiInDevice::Close ( )
{
if (!IsOpen())                          /* if no device open                 */
  return FALSE;                         /* return error                      */
bClosing = TRUE;                        /* set flag that we're closing now   */
Stop();                                 /* stop recording from device        */
Reset();                                /* reset the attached device         */
UnprepareBuf();                         /* unprepare allocated data          */

midiInClose(hDev);                      /* close the attached device         */
hDev = NULL;                            /* and reset the handle              */
bClosing = FALSE;
return CMidiDevice::Close();            /* and pass back OK                  */
}

/*****************************************************************************/
/* Message : sends a special message to the device                           */
/*****************************************************************************/

DWORD CMidiInDevice::Message
    (
    UINT msg,
    DWORD dw1,
    DWORD dw2
    )
{
DWORD rc = 0;

if (hDev)
  rc = midiInMessage(hDev, msg, dw1, dw2);
return rc;
}

/*****************************************************************************/
/* IsRecording : returns whether currently recording                         */
/*****************************************************************************/

BOOL CMidiInDevice::IsRecording ( )
{
return bRecording;
}

/*****************************************************************************/
/* AddHeader : adds a MIDI Input buffer to the device                        */
/*****************************************************************************/

BOOL CMidiInDevice::AddHeader
    (
    LPMIDIHDR lpHdr
    )
{
if (lpHdr->dwFlags & MHDR_PREPARED)
#if 1
  return TRUE;
#else
  {
  UnprepareHeader(lpHdr);
  lpHdr->dwBufferLength = MIDIINBUFSIZE;
  }
#endif
if (!PrepareHeader(lpHdr))
  return(FALSE);
if (midiInAddBuffer(hDev, lpHdr, sizeof(MIDIHDR)))
  return(FALSE);
return(TRUE);
}

/*****************************************************************************/
/* CheckHeader : checks the passed MIDI header for validity                  */
/*****************************************************************************/

BOOL CMidiInDevice::CheckHeader
    (
    LPMIDIHDR lpHdr
    )
{
int i;

if ((IsBadReadPtr(lpHdr, sizeof(MIDIHDR))) ||
    (!lpGlHdr))
  return FALSE;
for (i=0; i < MIDIINHDRS; i++)          /* if it's one we allocated          */
  if (lpHdr == lpGlHdr[i])
    return TRUE;                        /* pass back OK                      */
return FALSE;
}

/*****************************************************************************/
/* AllocBuf : allocates MIDI In buffers                                      */
/*****************************************************************************/

BOOL CMidiInDevice::AllocBuf ( )
{
int i;                                  /* loop counter                      */

if (hGlobal)                            /* if already allocated              */
  return TRUE;                          /* pass back OK                      */

hGlobal = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,
    MIDIINHDRS * (sizeof(MIDIHDR) + MIDIINBUFSIZE));
if (hGlobal == NULL)
  return FALSE;

pGlobal = (char *)GlobalLock(hGlobal);  /* lock allocated memory             */
if ((pGlobal == NULL) ||
    (lpGlHdr == NULL))
  {
  GlobalFree(hGlobal);
  hGlobal = NULL;
  return FALSE;
  }

for (i=0; i < MIDIINHDRS; i++)          /* now generate the MIDI buffers     */
  {
  lpGlHdr[i] = (LPMIDIHDR)
      (pGlobal + i * (sizeof(MIDIHDR) + MIDIINBUFSIZE));
  memset(lpGlHdr[i], 0, sizeof(MIDIHDR) + MIDIINBUFSIZE);
  lpGlHdr[i]->dwBufferLength = MIDIINBUFSIZE;
  lpGlHdr[i]->lpData = (char *)(lpGlHdr[i]+1);
  }

PrepareBuf();                           /* and prepare them for usage        */
return TRUE;
}

/*****************************************************************************/
/* FreeBuf : frees allocated MIDI In buffers                                 */
/*****************************************************************************/

BOOL CMidiInDevice::FreeBuf ( )
{
int i;

if (!hGlobal)                           /* if nothing allocated              */
  return(TRUE);                         /* pass back OK                      */

UnprepareBuf();                         /* make sure everything's unprepared */

for (i=0; i < MIDIINHDRS; i++)          /* otherwise remove the MIDI buffers */
  lpGlHdr[i] = NULL;                    /* reset buffer pointer              */

pGlobal = NULL;                         /* reset global buffer pointer       */
GlobalUnlock(hGlobal);                  /* unlock the global buffer          */
GlobalFree(hGlobal);                    /* free it                           */
hGlobal = NULL;                         /* and reset the handle              */
return(TRUE);
}

/*****************************************************************************/
/* PrepareBuf : prepares MIDI In buffers                                     */
/*****************************************************************************/

BOOL CMidiInDevice::PrepareBuf ( )
{
int i;                                  /* loop counter                      */

if ((!hGlobal) || (!pGlobal))           /* if not already allocated          */
  return FALSE;                         /* pass back NOT OK                  */

for (i=0; i < MIDIINHDRS; i++)          /* now prepare the MIDI buffers      */
  if (!AddHeader(lpGlHdr[i]))
    return FALSE;

return TRUE;
}

/*****************************************************************************/
/* UnprepareBuf : unprepares MIDI In buffers                                 */
/*****************************************************************************/

BOOL CMidiInDevice::UnprepareBuf ( )
{
int i;

if ((!hGlobal)||(!pGlobal))             /* if nothing allocated              */
  return(TRUE);                         /* pass back OK                      */

for (i=0; i < MIDIINHDRS; i++)          /* otherwise remove the MIDI buffers */
  {
  if ((lpGlHdr[i]) &&
      (lpGlHdr[i]->dwFlags & MHDR_PREPARED))
    UnprepareHeader(lpGlHdr[i]);
  }

return(TRUE);
}

/*****************************************************************************/
/* PreparedBuffers : returns # buffers that still are prepared               */
/*****************************************************************************/

int CMidiInDevice::PreparedBuffers()
{
int i, prep = 0;

if ((!hGlobal)||(!pGlobal))             /* if nothing allocated              */
  return 0;                             /* pass back NONE                    */

for (i=0; i < MIDIINHDRS; i++)          /* otherwise remove the MIDI buffers */
  {
  if ((lpGlHdr[i]) &&
      (lpGlHdr[i]->dwFlags & MHDR_PREPARED))
    prep++;
  }

return prep;
}

/*****************************************************************************/
/* OnMimData : reacts on incoming MM_MIM_DATA message                        */
/*****************************************************************************/

LRESULT CMidiInDevice::OnMimData
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
CMidiMsg iMsg((DWORD)lParam);           /* generate MIDI message             */
iMsg.SetStamp(::timeGetTime() - dwStamp);
// TRACE2("MIDI IN at %d (%d)\n", iMsg.GetStamp(), dwStamp);
Data(iMsg);                             /* process the message               */
return 0;
}

/*****************************************************************************/
/* OnMimLongData : reacts on incoming MM_MIM_LONGDATA message                */
/*****************************************************************************/

void CMidiInDevice::OnMimLongData(LPMIDIHDR lphdr, DWORD dwTS)
{
if ((!lphdr) ||                         /* if erroneous or invalid pointer   */
    (!CheckHeader(lphdr)))
  return;                               /* treat as if no error happened     */

UnprepareHeader(lphdr);
if (!bRecording)                        /* if recording stopped              */
  return;                               /* do nothing more                   */

if (lphdr->dwBytesRecorded)
  {
  CMidiMsg iMsg(lphdr);                 /* generate MIDI message             */
  iMsg.SetStamp(dwTS);
  Data(iMsg);                           /* process the message               */
  }

AddHeader(lphdr);                       /* then re-enable the buffer         */
}

LRESULT CMidiInDevice::OnMimLongData
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMimLongData((LPMIDIHDR)lParam, ::timeGetTime() - dwStamp);
return 0;
}

/*****************************************************************************/
/* OnMimError : called upon (MM_)MIM_(LONG)ERROR messages                    */
/*****************************************************************************/

void CMidiInDevice::OnMimError(LPMIDIHDR lphdr, DWORD dwTS)
{
if ((!lphdr) ||                         /* if erroneous or invalid pointer   */
    (!CheckHeader(lphdr)))
  return;                               /* treat as if no error happened     */

UnprepareHeader(lphdr);
if (!bRecording)                        /* if recording stopped              */
  return;                               /* do nothing more                   */

if (lphdr->dwBytesRecorded)
  {
  CMidiMsg iMsg(lphdr);                 /* generate MIDI message             */
  iMsg.SetStamp(dwTS ? dwTS : ::timeGetTime() - dwStamp);
  OnMimError(iMsg);
  }
AddHeader(lphdr);                       /* then re-enable the buffer         */
}

LRESULT CMidiInDevice::OnMimError
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
CMidiMsg iMsg((DWORD)lParam);           /* generate MIDI message             */
iMsg.SetStamp(::timeGetTime() - dwStamp);
OnMimError(iMsg);                       /* process the message               */
return 0;
}

LRESULT CMidiInDevice::OnMimLongError
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMimError((LPMIDIHDR)lParam, ::timeGetTime() - dwStamp);
return 0;
}

/*****************************************************************************/
/* OnMimOpen : called when the device is opened                              */
/*****************************************************************************/

LRESULT CMidiInDevice::OnMimOpen
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMimOpen();
return 0;
}

/*****************************************************************************/
/* OnMimClose : called when the device is closed                             */
/*****************************************************************************/

LRESULT CMidiInDevice::OnMimClose
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMimClose();
return 0;
}

/*****************************************************************************/
/* OnMimMoreData : called when a MM_MIM_MOREDATA message is processed        */
/*****************************************************************************/

LRESULT CMidiInDevice::OnMimMoreData
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
CMidiMsg iMsg((DWORD)lParam);           /* generate MIDI message             */
iMsg.SetStamp(::timeGetTime() - dwStamp);
OnMimMoreData(iMsg);
return 0;
}

/*****************************************************************************/
/* Data : processes incoming MIDI data                                       */
/*****************************************************************************/

void CMidiInDevice::Data
    (
    CMidiMsg & Msg
    )
{
switch (Msg[0] & 0xf0)                  /* look what kind of message :       */
  {
  case STATUS_NOTEOFF :
    NoteOff(Msg[0] & 0x0f, Msg[1]);
    break;
  case STATUS_NOTEON :
    NoteOn(Msg[0] & 0x0f, Msg[1], Msg[2]);
    break;
  case STATUS_POLYPHONICKEY :
    KeyAftertouch(Msg[0] & 0x0f, Msg[1], Msg[2]);
    break;
  case STATUS_CONTROLCHANGE :
    Controller(Msg[0] & 0x0f, Msg[1], Msg[2]);
    break;
  case STATUS_PROGRAMCHANGE :
    Patch(Msg[0] & 0x0f, Msg[1]);
    break;
  case STATUS_CHANNELPRESSURE :
    ChannelAftertouch(Msg[0] & 0x0f, Msg[1]);
    break;
  case STATUS_PITCHBEND :
    PitchBend(Msg[0] & 0x0f, (int)Msg[1] | ((int)Msg[2] << 7));
    break;
  case STATUS_SYSTEMRESET :
    SystemReset();
    break;
  }
}

/*===========================================================================*/
/* Class CMidiOutDeviceList                                                  */
/*===========================================================================*/

/*****************************************************************************/
/* CMidiOutDeviceList : constructor                                          */
/*****************************************************************************/

CMidiOutDeviceList::CMidiOutDeviceList ( )
{
WORD wNumDevices;
int i;
MIDIOUTCAPS mcap;

Add(szMMapName);
wNumDevices = midiOutGetNumDevs();      /* get # MIDI devices                */
if (wNumDevices)
  {
  for (i=0; i < (int) wNumDevices; i++)
    {
    midiOutGetDevCaps(i,
        &mcap,
        sizeof(mcap));
    Add(mcap.szPname);
    }
  }
}

/*===========================================================================*/
/* Class CMidiOutDevice                                                      */
/*===========================================================================*/

BEGIN_MESSAGE_MAP(CMidiOutDevice, CMidiDevice)
    //{{AFX_MSG_MAP(CMidiOutDevice)
    ON_MESSAGE(MM_MOM_OPEN, OnMomOpen)
    ON_MESSAGE(MM_MOM_DONE, OnMomDone)
    ON_MESSAGE(MM_MOM_CLOSE, OnMomClose)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*****************************************************************************/
/* CMidiOutDevice : constructor                                              */
/*****************************************************************************/

CMidiOutDevice::CMidiOutDevice ( )
{
hDev = NULL;                            /* reset device handle               */
RunningStatus = 0;                      /* reset running status              */
}

/*****************************************************************************/
/* ~CMidiOutDevice : destructor                                              */
/*****************************************************************************/

CMidiOutDevice::~CMidiOutDevice ( )
{
if (IsOpen())
  Close();
}

/*****************************************************************************/
/* MidiOutProc : callback procedure                                          */
/*****************************************************************************/

void CALLBACK CMidiOutDevice::MidiOutProc
    (
    HMIDIOUT hMidiOut,
    UINT wMsg,
    DWORD dwInstance,
    DWORD dwParam1,
    DWORD dwParam2
    )
{
CMidiOutDevice * pOutDev = (CMidiOutDevice *)dwInstance;
switch (wMsg)
  {
  case MOM_OPEN : 
    pOutDev->OnMomOpen();
    break;
  case MOM_DONE : 
    pOutDev->OnMomDone((LPMIDIHDR)dwParam1);
    break;
  case MOM_CLOSE : 
    pOutDev->OnMomClose();
    break;
  }
}

/*****************************************************************************/
/* Open : opens a MIDI In Device with its ID                                 */
/*****************************************************************************/

BOOL CMidiOutDevice::Open
    (
    int iID,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
if (IsOpen())                           /* if already open                   */
  Close();                              /* close the opened device           */

AssureWnd();                            /* make sure hWnd is there           */

if ((dwFlags & CALLBACK_TYPEMASK) ==    /* if function callback              */
    CALLBACK_FUNCTION)
  {
  if (lp1 == OPENDEFAULT)               /* if default value                  */
    lp1 = (DWORD)MidiOutProc;           /* set LP1 to procedure              */
  if (lp2 == OPENDEFAULT)               /* if default value                  */
    lp2 = (DWORD)this;                  /* set LP2 to object                 */
  }
else
  {
  if (lp1 == OPENDEFAULT)               /* if default value                  */
    lp1 = (UINT)m_hWnd;                 /* set LP1 to window                 */
  if (lp2 == OPENDEFAULT)               /* if default value                  */
    lp2 = (UINT)AfxGetInstanceHandle(); /* set LP2 to instance handle        */
  }

WORD wRtn = 1;
try
  {
  wRtn = midiOutOpen(&hDev,             /* open the device                   */
      iID,                              /* with the passed ID                */
      lp1, lp2, dwFlags);
  }
catch(...)
  {
  }
if (wRtn)                               /* if open error                     */
  {
  hDev = NULL;                          /* make sure that handle is reset    */
  return FALSE;
  }

midiOutGetDevCaps(iID,                  /* get device capabilities           */
    &mCap, sizeof(mCap));

return CMidiDevice::Open(iID,           /* pass back OK                      */
    lp1, lp2, dwFlags);
}

/*****************************************************************************/
/* Open : opens a device with its name                                       */
/*****************************************************************************/

BOOL CMidiOutDevice::Open
    (
    LPCSTR szName,
    DWORD lp1,
    DWORD lp2,
    DWORD dwFlags
    )
{
WORD wNumDevices;
int i;
WORD wRtn;
MIDIOUTCAPS mcap;

wNumDevices = midiOutGetNumDevs();      /* get # MIDI input devices          */
if (wNumDevices)
  {
  for (i=-1; i < (int) wNumDevices; i++)
    {
    if (i == -1)
      lstrcpy(mcap.szPname, szMMapName);
    else
      wRtn = midiOutGetDevCaps(i,
          &mcap,
          sizeof(mcap));
    if (!lstrcmp(szName, mcap.szPname))
      {
      if (!Open(i, lp1, lp2, dwFlags))
        return(FALSE);                  /* error - return NOT OK             */

      return CMidiDevice::Open(szName,  /* opened - return OK                */
          lp1, lp2, dwFlags );
      }
    }
  }
return(FALSE);                          /* return NOT OK                     */
}

/*****************************************************************************/
/* Reset : resets the attached MIDI In device                                */
/*****************************************************************************/

BOOL CMidiOutDevice::Reset ( )
{
if (!IsOpen())                          /* if no device open                 */
  return FALSE;                         /* return error                      */
midiOutReset(hDev);                     /* reset the attached device         */
return TRUE;                            /* and pass back OK                  */
}

/*****************************************************************************/
/* Close : closes the opened MIDI In Device                                  */
/*****************************************************************************/

BOOL CMidiOutDevice::Close ( )
{
if (!IsOpen())                          /* if no device open                 */
  return FALSE;                         /* return error                      */
Reset();                                /* reset the attached device         */
midiOutClose(hDev);                     /* close the attached device         */
hDev = NULL;                            /* and reset the handle              */
return CMidiDevice::Close();            /* and pass back OK                  */
}

/*****************************************************************************/
/* OnMomOpen : reacts on incoming MM_MOM_OPEN message                        */
/*****************************************************************************/

LRESULT CMidiOutDevice::OnMomOpen
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMomOpen();
return 0;
}

/*****************************************************************************/
/* OnMomDone : reacts on incoming MM_MOM_DONE message                        */
/*****************************************************************************/

LRESULT CMidiOutDevice::OnMomDone
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMomDone((LPMIDIHDR)lParam);
return 0;
}

/*****************************************************************************/
/* OnMomClose : reacts on incoming MM_MOM_CLOSE message                      */
/*****************************************************************************/

LRESULT CMidiOutDevice::OnMomClose
    (
    WPARAM wParam,
    LPARAM lParam
    )
{
OnMomClose();
return 0;
}

/*****************************************************************************/
/* Message : sends a special message to the device                           */
/*****************************************************************************/

DWORD CMidiOutDevice::Message
    (
    UINT msg,
    DWORD dw1,
    DWORD dw2
    )
{
DWORD rc = 0;

if (hDev)
  rc = midiOutMessage(hDev, msg, dw1, dw2);
return rc;
}

/*****************************************************************************/
/* SendingDone : returns whether buffer has been sent                        */
/*****************************************************************************/

BOOL CMidiOutDevice::SendingDone
    (
    LPMIDIHDR lpHdr
    )
{
return((lpHdr->dwFlags & MHDR_DONE) ? TRUE : FALSE);
}

/*****************************************************************************/
/* CachePatches : caches MIDI device patches                                 */
/*****************************************************************************/

BOOL CMidiOutDevice::CachePatches
    (
    UINT uBank,
    WORD FAR* lpwPatchArray,
    UINT uFlags
    )
{
if ((!IsOpen()) ||                      /* if not opened / supported         */
    (!(mCap.dwSupport & MIDICAPS_CACHE)))
  return FALSE;                         /* pass back error                   */

return !midiOutCachePatches(hDev,
    uBank, lpwPatchArray, uFlags);
}

/*****************************************************************************/
/* CacheDrumPatches : caches drum patches                                    */
/*****************************************************************************/

BOOL CMidiOutDevice::CacheDrumPatches
    (
    UINT uPatch,
    WORD FAR* lpwKeyArray,
    UINT uFlags
    )
{
if ((!IsOpen()) ||                      /* if not opened / supported         */
    (!(mCap.dwSupport & MIDICAPS_CACHE)))
  return FALSE;                         /* pass back error                   */

return !midiOutCacheDrumPatches(hDev,
  uPatch, lpwKeyArray, uFlags);
}

/*****************************************************************************/
/* Output : sends MIDI data to output                                        */
/*****************************************************************************/

BOOL CMidiOutDevice::Output
    (
    CMidiMsg &Msg
    )
{
if (!IsOpen())                          /* if not opened                     */
  return FALSE;                         /* pass back error                   */

switch (Msg.Length())                   /* which length has the message ?    */
  {
  case 0 :                              /* none ?                            */
    return FALSE;                       /* pass back error                   */
  case 1 :                              /* short message ?                   */
  case 2 :
  case 3 :
    {
    BYTE bRS = Msg[0] & 0xf0;           /* get running status                */
    BYTE bCh = Msg[0] & 0x0f;
    if ((bRS == 0x80) &&                /* if sending NOTE OFF on running chn*/
        (RunningStatus == (0x90|bCh)))  /* and running status is NOTE ON     */
      {
      Msg[0] = 0x90 | bCh;              /* make it a NOTE ON                 */
      Msg[2] = 0;                       /* with velocity 0                   */
      }
    if (bRS < 0xf0)                     /* if no realtime / system message   */
      RunningStatus = Msg[0];           /* remember running status           */
    return !midiOutShortMsg(hDev, Msg);
    }
  default :
    {
    UINT rc = midiOutPrepareHeader(hDev, Msg, sizeof(MIDIHDR));
    if (!rc)
      {
      rc = midiOutLongMsg(hDev, Msg, sizeof(MIDIHDR));
      while ((!rc) && (!SendingDone(Msg)))
        ;
      midiOutUnprepareHeader(hDev, Msg, sizeof(MIDIHDR));
      }
    return !rc;
    }
  }
}

/*****************************************************************************/
/* NoteOn : sends a NOTE ON to output                                        */
/*****************************************************************************/

BOOL CMidiOutDevice::NoteOn
    (
    int channel,
    int note,
    int vel
    )
{
CMidiMsg msg(STATUS_NOTEON | (channel & 0x0f), note, vel);
return Output(msg);
}

/*****************************************************************************/
/* NoteOff : sends a NOTE OFF to output                                      */
/*****************************************************************************/

BOOL CMidiOutDevice::NoteOff
    (
    int channel,
    int note,
    int vel
    )
{
// this is eventually corrected to STATUS_NOTEON with vel 0 in 
// Output() if the running status is NOTE ON
CMidiMsg msg(STATUS_NOTEOFF | (channel & 0x0f), note, vel);
return Output(msg);
}

/*****************************************************************************/
/* KeyAftertouch : sends a Key Aftertouch to output                          */
/*****************************************************************************/

BOOL CMidiOutDevice::KeyAftertouch
    (
    int channel,
    int note,
    int vel
    )
{
CMidiMsg msg(STATUS_POLYPHONICKEY | (channel & 0x0f), note, vel);
return Output(msg);
}

/*****************************************************************************/
/* ChannelAftertouch : sends a channel aftertouch to output                  */
/*****************************************************************************/

BOOL CMidiOutDevice::ChannelAftertouch
    (
    int channel,
    int vel
    )
{
CMidiMsg msg(STATUS_CHANNELPRESSURE | (channel & 0x0f), vel);
return Output(msg);
}

/*****************************************************************************/
/* Controller : sends a 7-bit controller to output                           */
/*****************************************************************************/

BOOL CMidiOutDevice::Controller
    (
    int channel,
    int ctrl,
    int vel
    )
{
CMidiMsg msg(STATUS_CONTROLCHANGE | (channel & 0x0f), ctrl, vel);
return Output(msg);
}

/*****************************************************************************/
/* Controller14 : sends a 14-bit controller to output                        */
/*****************************************************************************/

BOOL CMidiOutDevice::Controller14
    (
    int channel,
    int ctrl,
    int vel
    )
{
CMidiMsg msg(STATUS_CONTROLCHANGE | (channel & 0x0f), ctrl, (vel >> 7));
CMidiMsg msg2(STATUS_CONTROLCHANGE | (channel & 0x0f), ctrl+0x20, (vel & 0x7f));
if (Output(msg))
  return Output(msg2);
return FALSE;
}

/*****************************************************************************/
/* Patch : sends a patch change to output                                    */
/*****************************************************************************/

BOOL CMidiOutDevice::Patch
    (
    int channel,
    int patch
    )
{
CMidiMsg msg(STATUS_PROGRAMCHANGE | (channel & 0x0f), patch);
return Output(msg);
}

/*****************************************************************************/
/* PitchBend : sends pitch bend data to output                               */
/*****************************************************************************/

BOOL CMidiOutDevice::PitchBend
    (
    int channel,
    int val
    )
{
CMidiMsg msg(STATUS_PITCHBEND | (channel & 0x0f), (val & 0x7f), (val >> 7));
return Output(msg);
}

/*****************************************************************************/
/* SystemReset : sends system reset message to output                        */
/*****************************************************************************/

BOOL CMidiOutDevice::SystemReset ( void )
{
CMidiMsg msg((DWORD)STATUS_SYSTEMRESET);
return Output(msg);
}

/*****************************************************************************/
/* Bank : activates another bank on the device                               */
/*****************************************************************************/

BOOL CMidiOutDevice::Bank
    (
    int channel, 
    int bank
    )
{ 
return Controller(channel, 0, bank);
};

/*****************************************************************************/
/* Bank14 : activates another bank on the device with 14-bit resolution      */
/*****************************************************************************/

BOOL CMidiOutDevice::Bank14
    (
    int channel, 
    int bank
    )
{
return Controller14(channel, 0, bank);
};
