/*****************************************************************************/
/* MFCMIDI.H : class definition for MIDI base classes                        */
/*****************************************************************************/
/*                                                                           */
/* $Archive::                                                              $ */
/* $Revision::                                                             $ */
/* $Date::                                                                 $ */
/* $Author::                                                               $ */
/* $Modtime::                                                              $ */
/* $Workfile::                                                             $ */
/*                                                                           */
/*****************************************************************************/

#ifndef __INCLUDE_MFCMIDI_H__
#define __INCLUDE_MFCMIDI_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#ifndef __cplusplus
#error mfcmidi.h is for use with C++
#endif

/*****************************************************************************/
/* Necessary includes                                                        */
/*****************************************************************************/

/*****************************************************************************/
/* MIDI definitions                                                          */
/*****************************************************************************/

#define ISSTATUS(bData)         ((bData) & 0x80)
#define FILTERCHANNEL(bStatus)  ((BYTE)((bStatus) & 0xf0))
#define FILTERSTATUS(bStatus)   ((BYTE)((bStatus) & 0x0f))

#define STATUS_NOTEOFF          0x80
#define STATUS_NOTEON           0x90
#define STATUS_POLYPHONICKEY    0xa0
#define STATUS_CONTROLCHANGE    0xb0
#define STATUS_PROGRAMCHANGE    0xc0
#define STATUS_CHANNELPRESSURE  0xd0
#define STATUS_PITCHBEND        0xe0

#define STATUS_SYS              0xf0
#define STATUS_SYSEX            0xf0
#define STATUS_QFRAME           0xf1
#define STATUS_SONGPOINTER      0xf2
#define STATUS_SONGSELECT       0xf3
#define STATUS_F4               0xf4
#define STATUS_F5               0xf5
#define STATUS_TUNEREQUEST      0xf6
#define STATUS_EOX              0xf7
#define STATUS_TIMINGCLOCK      0xf8
#define STATUS_F9               0xf9
#define STATUS_START            0xfa
#define STATUS_CONTINUE         0xfb
#define STATUS_STOP             0xfc
#define STATUS_FD               0xfd
#define STATUS_ACTIVESENSING    0xfe
#define STATUS_SYSTEMRESET      0xff

#define OPENDEFAULT             ((DWORD)-1)

/*****************************************************************************/
/* CMidiMsg : a MIDI message                                                 */
/*****************************************************************************/

class CMidiMsg
  {
  public:
    CMidiMsg(DWORD smsg = 0);
    CMidiMsg(int b1, int b2, int b3=0);
    CMidiMsg(LPMIDIHDR lphdr);
    CMidiMsg(LPSTR data, int len);
    CMidiMsg (CMidiMsg const & org)
      { Init(); DoCopy(org); }
//    virtual ~CMidiMsg()               /* virtual destructor not needed     */
                                        /* as there are no derived classes   */
    ~CMidiMsg()                         /* destructor                        */
       { Empty(); }
                                        /* assignment operator               */
    CMidiMsg & operator=(CMidiMsg const &org) 
      { return DoCopy(org); }
    CMidiMsg & operator=(DWORD smsg)
      { Set(smsg); return *this; }
    CMidiMsg & operator=(LPMIDIHDR lphdr)
      { Set(lphdr); return *this; }
                                        /* set to MIDI message               */
    void Set(DWORD smsg = 0);
    void Set(LPMIDIHDR lphdr);
    void Set(LPSTR data, int len);
    void Set(int b1, int b2, int b3);
    void SetStamp(DWORD dwTS = 0) { dwStamp = dwTS; }
                                        /* convert to other data             */
    operator char*();
    operator BYTE*();
    operator DWORD();
    operator LPMIDIHDR();
                                        /* get x-th byte of message          */
    BYTE & operator [](int idx) const
      { return ((CMidiMsg *)this)->GetAt(idx); }
    BYTE & GetAt(int idx)
      {
      if (!lpHdr)
        return ((BYTE *)&sMsg)[idx];
      return ((BYTE *)lpHdr->lpData)[idx];
      }
    WORD Length();                      /* get message length                */
                                        /* get timestamp                     */
    DWORD GetStamp() { return dwStamp; }
  protected:
    BOOL bAlloc;                        /* flag whether self-allocated buffer*/
    DWORD sMsg;                         /* short message                     */
    LPMIDIHDR lpHdr;                    /* long message header               */
    HGLOBAL hGlPtr;                     /* global pointer                    */
    DWORD dwStamp;                      /* time stamp                        */
    void Init();                        /* initialize memory                 */
    void Empty();                       /* clear memory                      */
                                        /* copy another MIDI message         */
    CMidiMsg & DoCopy(CMidiMsg const &org)
      {
      Empty();                          /* remove any previous contents      */

      sMsg = org.sMsg;                  /* store short message               */
      if (org.lpHdr)                    /* if original has a buffer          */
                                        /* copy original's buffer            */
        Set(org.lpHdr->lpData, (int)org.lpHdr->dwBufferLength);
      dwStamp = org.dwStamp;
      return *this;                     /* pass back pointer to ourselves    */
      }
      
  };

/*****************************************************************************/
/* CMidiDevice : MIDI Device base class                                      */
/*****************************************************************************/

class CMidiDevice : public CWnd
  {
// Attributes
  public:
    CMidiDevice();                      /* constructor                       */
    ~CMidiDevice();                     /* destructor                        */

    BOOL IsOpen()                       /* return whether device is open     */
      { return bIsOpen; }
    LPCSTR DeviceName ( )               /* return opened device' name        */
      { return bIsOpen ? szDevName : ""; }

// Operations
     virtual BOOL Open(LPCSTR szName,   /* open with device name             */
         DWORD lp1 = OPENDEFAULT,
         DWORD lp2 = OPENDEFAULT,
         DWORD dwFlags = CALLBACK_WINDOW);
     virtual BOOL Open(int iID,         /* open with device ID               */
         DWORD lp1 = OPENDEFAULT,
         DWORD lp2 = OPENDEFAULT,
         DWORD dwFlags = CALLBACK_WINDOW);
     virtual void OnOpen() {}
     virtual BOOL Reset();              /* reset device                      */
     virtual BOOL Close();              /* close device                      */
     virtual void OnClose() {}
     virtual DWORD Message(UINT msg,    /* send special message              */
                           DWORD dw1,
                           DWORD dw2)
       { return 0; }

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CMidiDevice)
	//}}AFX_VIRTUAL

	// Generated message map functions
protected:
	//{{AFX_MSG(CMidiDevice)
		// NOTE - the ClassWizard will add and remove member functions here.
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()


// Implementation
  private :
    BOOL bIsOpen;                       /* flag whether opened               */

  protected :
    BOOL Create();
    void AssureWnd()
      {
      if (!m_hWnd)                      /* if no notification window yet     */
        {
        Create();                       /* create it                         */
        ASSERT(m_hWnd);
        }
      }

    char szDevName[MAXPNAMELEN];        /* device name                       */
  };

/*****************************************************************************/
/* CMidiInDeviceList : MIDI Input device list                                */
/*****************************************************************************/

class CMidiInDeviceList : public CStringArray
  {
  public :
    CMidiInDeviceList ( );              /* constructor                       */
  };

/*****************************************************************************/
/* CMidiInDevice : MIDI Input device                                         */
/*****************************************************************************/

class CMidiInDevice : public CMidiDevice
  {
// Attributes
  public:
    BOOL IsRecording();

// Operations
public:
    CMidiInDevice();
    virtual BOOL Open(LPCSTR szName,    /* open with device name             */
        DWORD lp1 = OPENDEFAULT,
        DWORD lp2 = OPENDEFAULT,
        DWORD dwFlags = CALLBACK_WINDOW);
    virtual BOOL Open(int iID,          /* open with device ID               */
        DWORD lp1 = OPENDEFAULT,
        DWORD lp2 = OPENDEFAULT,
        DWORD dwFlags = CALLBACK_WINDOW);
                                        /* prepare LPMIDIHDR                 */
    virtual BOOL PrepareHeader(LPMIDIHDR lpHdr);
                                        /* unprepare LPMIDIHDR               */
    virtual BOOL UnprepareHeader(LPMIDIHDR lpHdr);
    virtual BOOL Start();               /* start recording                   */
    virtual BOOL Stop();                /* stop recording                    */
    virtual BOOL Reset();               /* reset device                      */
    virtual BOOL Close();               /* close device                      */
    virtual DWORD Message(UINT msg,     /* send special message              */
        DWORD dw1, DWORD dw2);
                                        /* process incoming data             */
    virtual void Data(CMidiMsg &Msg);
    virtual void OnMimOpen() { }
    virtual void OnMimClose() { }
    virtual void OnMimError(CMidiMsg &Msg) { }
    virtual void OnMimError(LPMIDIHDR lphdr, DWORD dwTS = 0);
    virtual void OnMimLongData(LPMIDIHDR lphdr, DWORD dwTS = 0);
    virtual void OnMimMoreData(CMidiMsg &Msg) { }
    virtual void NoteOn(int channel, int note, int vel) { }
    virtual void NoteOff(int channel, int note, int vel = 64) { }
    virtual void KeyAftertouch(int channel, int note, int vel) { }
    virtual void ChannelAftertouch(int channel, int vel) { }
    virtual void Controller(int channel, int ctrl, int val) { }
    virtual void Patch(int channel, int patch) { }
    virtual void PitchBend(int channel, int val) { }
    virtual void SysEx(CMidiMsg &msg) { }
    virtual void SystemReset() { };

// Implementation
public:
    virtual ~CMidiInDevice();

protected:
    HMIDIIN hDev;                       /* input device handle               */
    HGLOBAL hGlobal;                    /* global memory for SysEx buffers   */
    char * pGlobal;                     /* same as locked pointer            */
    LPMIDIHDR *lpGlHdr;                 /* MIDI input buffers                */
    BOOL bRecording;                    /* flag whether currently recording  */
    DWORD dwStamp;                      /* internal start time stamp         */
    MIDIINCAPS mCap;                    /* input device capabilities         */
    BOOL bClosing;                      /* flag whether currently closing    */

protected:
                                        /* add MIDI buffer to device         */
    BOOL AddHeader(LPMIDIHDR lpHdr);
                                        /* check whether it's one of ours    */
    BOOL CheckHeader(LPMIDIHDR lpHdr);  
    BOOL AllocBuf();                    /* allocate buffer                   */
    BOOL FreeBuf();                     /* free allocated buffer             */
    BOOL PrepareBuf();                  /* prepare buffer for usage          */
    BOOL UnprepareBuf();                /* unprepare buffer from usage       */
    int PreparedBuffers();              /* # currently prepared buffers      */

    static void CALLBACK MidiInProc
      (
      HMIDIIN hMidiIn,
      UINT wMsg,
      DWORD dwInstance,
      DWORD dwParam1,
      DWORD dwParam2
      );
public:
// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CMidiInDevice)
	//}}AFX_VIRTUAL

    // Generated message map functions
protected:
    //{{AFX_MSG(CMidiInDevice)
    afx_msg LRESULT OnMimOpen(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimClose(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimError(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimLongError(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimData(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimLongData(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMimMoreData(WPARAM w, LPARAM l);
    //}}AFX_MSG
    DECLARE_MESSAGE_MAP()
  };

/*****************************************************************************/
/* CMidiOutDeviceList : MIDI Output device list                              */
/*****************************************************************************/

class CMidiOutDeviceList : public CStringArray
  {
  public :
    CMidiOutDeviceList ( );             /* constructor                       */
  };

/*****************************************************************************/
/* CMidiOutDevice : MIDI Output device                                       */
/*****************************************************************************/

class CMidiOutDevice : public CMidiDevice
  {
// Attributes
public:

// Operations
public:
    CMidiOutDevice();
    virtual BOOL Open(LPCSTR szName,    /* open with device name             */
        DWORD lp1 = OPENDEFAULT,
        DWORD lp2 = OPENDEFAULT,
        DWORD dwFlags = CALLBACK_WINDOW);
    virtual BOOL Open(int iID,          /* open with device ID               */
        DWORD lp1 = OPENDEFAULT,
        DWORD lp2 = OPENDEFAULT,
        DWORD dwFlags = CALLBACK_WINDOW);
    virtual BOOL Reset();
    virtual BOOL Close();
    virtual void OnMomClose() {}
    virtual void OnMomDone(LPMIDIHDR lphdr) {}
    virtual void OnMomOpen() {}
    virtual DWORD Message(UINT msg, DWORD dw1, DWORD dw2);
    virtual BOOL CachePatches(UINT uBank,
        WORD FAR* lpwPatchArray, UINT uFlags);
    virtual BOOL CacheDrumPatches(UINT uPatch,
        WORD FAR* lpwKeyArray, UINT uFlags);
    virtual BOOL Output(CMidiMsg &Msg);
    virtual BOOL NoteOn(int channel, int note, int vel);
    virtual BOOL NoteOff(int channel, int note, int vel = 64);
    virtual BOOL KeyAftertouch(int channel, int note, int vel);
    virtual BOOL ChannelAftertouch(int channel, int vel);
    virtual BOOL Controller(int channel, int ctrl, int val);
    virtual BOOL Controller14(int channel, int ctrl, int val);
    virtual BOOL Patch(int channel, int patch);
    virtual BOOL PitchBend(int channel, int val);
    virtual BOOL SysEx(CMidiMsg &msg) { return Output(msg); }
    virtual BOOL SystemReset();
    virtual BOOL Bank(int channel, int bank);
    virtual BOOL Bank14(int channel, int bank);

// Implementation
public:
    virtual ~CMidiOutDevice();

protected:
    HMIDIOUT hDev;                      /* output device handle              */
    MIDIOUTCAPS mCap;                   /* output device capabilities        */
    BYTE RunningStatus;                 /* running status                    */

// Overrides
public:
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CMidiOutDevice)
	//}}AFX_VIRTUAL

    // Generated message map functions
protected:
    //{{AFX_MSG(CMidiOutDevice)
    afx_msg LRESULT OnMomOpen(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMomDone(WPARAM w, LPARAM l);
    afx_msg LRESULT OnMomClose(WPARAM w, LPARAM l);
    //}}AFX_MSG
    DECLARE_MESSAGE_MAP()
    BOOL SendingDone(LPMIDIHDR lpHdr);

    static void CALLBACK MidiOutProc
        (
        HMIDIOUT hMidiOut,
        UINT wMsg,
        DWORD dwInstance,
        DWORD dwParam1,
        DWORD dwParam2
        );
  };

#endif                                  /* __INCLUDE_MFCMIDI_H__             */
